@@ -1,21 +1,23 @@
 # encoding: utf-8
-import uuid
-
 from urllib.parse import urlencode
 from urllib.parse import urlparse
 
 import inspect
 import json
 import logging
-import six
 import sys
 import time
 
 from cryptojwt import as_unicode
-from cryptojwt import jws
+from cryptojwt.exception import UnsupportedAlgorithm
+from cryptojwt.jws.jws import factory as jws_factory
+from cryptojwt.jws.utils import left_hash
+from cryptojwt.jwt import JWT
 
 from oidcmsg import oauth2
 from oidcmsg import time_util
+from oidcmsg import verified_claim_name
+from oidcmsg.exception import FormatError
 from oidcmsg.exception import InvalidRequest
 from oidcmsg.exception import IssuerMismatch
 from oidcmsg.exception import MessageException
@@ -63,6 +65,22 @@ class IATError(VerificationError):
     pass
 
 
+def deserialize_from_one_of(val, msgtype, sformat):
+    if sformat in ["dict", "json"]:
+        flist = ['json', 'urlencoded']
+        if not isinstance(val, str):
+            val = json.dumps(val)
+    else:
+        flist = ['urlencoded', 'json']
+
+    for _format in flist:
+        try:
+            return msgtype().deserialize(val, _format)
+        except FormatError:
+            pass
+    raise FormatError('Unexpected format')
+
+
 def json_ser(val, sformat=None, lev=0):
     return json.dumps(val)
 
@@ -84,21 +102,11 @@ def idtoken_deser(val, sformat="urlencoded"):
 
 
 def address_deser(val, sformat="urlencoded"):
-    if sformat in ["dict", "json"]:
-        if not isinstance(val, six.string_types):
-            val = json.dumps(val)
-            sformat = "json"
-        elif sformat == "dict":
-            sformat = "json"
-    return AddressClaim().deserialize(val, sformat)
+    return deserialize_from_one_of(val, AddressClaim, sformat)
 
 
 def claims_deser(val, sformat="urlencoded"):
-    if sformat in ["dict", "json"]:
-        if not isinstance(val, six.string_types):
-            val = json.dumps(val)
-            sformat = "json"
-    return Claims().deserialize(val, sformat)
+    return deserialize_from_one_of(val, Claims, sformat)
 
 
 def msg_ser_json(inst, sformat="json", lev=0):
@@ -131,7 +139,7 @@ def msg_list_ser(insts, sformat, lev=0):
 
 def claims_ser(val, sformat="urlencoded", lev=0):
     # everything in c_extension
-    if isinstance(val, six.string_types):
+    if isinstance(val, str):
         item = val
     elif isinstance(val, list):
         item = val[0]
@@ -160,33 +168,35 @@ def claims_ser(val, sformat="urlencoded", lev=0):
 
 
 def registration_request_deser(val, sformat="urlencoded"):
-    if sformat in ["dict", "json"]:
-        if not isinstance(val, six.string_types):
-            val = json.dumps(val)
-            sformat = "json"
-    return RegistrationRequest().deserialize(val, sformat)
+    return deserialize_from_one_of(val, RegistrationRequest, sformat)
 
 
 def claims_request_deser(val, sformat="json"):
     # never 'urlencoded'
     if sformat == "urlencoded":
         sformat = "json"
-    if sformat in ["dict", "json"]:
-        if not isinstance(val, six.string_types):
+
+    if sformat == "json":
+        if not isinstance(val, str):
             val = json.dumps(val)
             sformat = "json"
+    elif sformat == 'dict':
+        if isinstance(val, str):
+            val = json.loads(val)
+
     return ClaimsRequest().deserialize(val, sformat)
 
 
 def dict_deser(val, sformat="json"):
-    # never 'urlencoded'
+    # never 'urlencoded', silently correct
     if sformat == "urlencoded":
         sformat = "json"
+
     if sformat in ["dict", "json"]:
-        if not isinstance(val, six.string_types):
+        if not isinstance(val, str):
             val = json.dumps(val)
-        elif isinstance(val, dict):
-            return val
+
+        return val
     else:
         raise ValueError('sformat can not be "{}"'.format(sformat))
 
@@ -229,9 +239,23 @@ def check_char_set(string, allowed):
 
 # -----------------------------------------------------------------------------
 
-ID_TOKEN_VERIFY_ARGS = ['keyjar','verify', 'encalg', 'encenc', 'sigalg',
+ID_TOKEN_VERIFY_ARGS = ['keyjar', 'verify', 'encalg', 'encenc', 'sigalg',
                         'issuer', 'allow_missing_kid', 'no_kid_issuer',
-                        'trusting', 'skew', 'nonce_storage_time', 'client_id']
+                        'trusting', 'skew', 'nonce_storage_time', 'client_id',
+                        'allow_sign_alg_none', 'allowed_sign_alg']
+
+CLAIMS_WITH_VERIFIED = ['id_token', 'id_token_hint', 'request']
+
+
+def clear_verified_claims(msg):
+    for claim in CLAIMS_WITH_VERIFIED:
+        _vc_name = verified_claim_name(claim)
+        try:
+            del msg[_vc_name]
+        except KeyError:
+            pass
+    return msg
+
 
 class RefreshAccessTokenRequest(oauth2.RefreshAccessTokenRequest):
     pass
@@ -241,26 +265,87 @@ class TokenErrorResponse(oauth2.TokenErrorResponse):
     pass
 
 
+def verify_id_token(msg, check_hash=False, claim='id_token', **kwargs):
+    # Try to decode the JWT, checks the signature
+    args = {}
+    for arg in ID_TOKEN_VERIFY_ARGS:
+        try:
+            args[arg] = kwargs[arg]
+        except KeyError:
+            pass
+
+    _jws = jws_factory(msg[claim])
+    if not _jws:
+        raise ValueError('{} not a signed JWT'.format(claim))
+
+    if _jws.jwt.headers['alg'] == 'none':
+        try:
+            _allow_none = kwargs['allow_sign_alg_none']
+        except KeyError:
+            logger.info('Signing algorithm None not allowed')
+            raise UnsupportedAlgorithm('Signing algorithm None not allowed')
+        else:
+            if not _allow_none:
+                logger.info('Signing algorithm None not allowed')
+                raise UnsupportedAlgorithm('Signing algorithm None not allowed')
+    else:
+        if "allowed_sign_alg" in kwargs:
+            if _jws.jwt.headers['alg'] != kwargs["allowed_sign_alg"]:
+                _msg = "Wrong token signing algorithm, {} != {}".format(
+                    _jws.jwt.headers['alg'], kwargs["allowed_sign_alg"])
+                logger.error(_msg)
+                raise UnsupportedAlgorithm(_msg)
+
+    _body = _jws.jwt.payload()
+    if 'keyjar' in kwargs:
+        try:
+            if _body['iss'] not in kwargs['keyjar']:
+                raise ValueError('Unknown issuer')
+        except KeyError:
+            raise MissingRequiredAttribute('iss')
+
+    idt = IdToken().from_jwt(str(msg[claim]), **args)
+    if not idt.verify(**kwargs):
+        return False
+
+    if check_hash:
+        _alg = idt.jws_header["alg"]
+        hfunc = "HS" + _alg[-3:]
+
+        if "access_token" in msg:
+            if "at_hash" not in idt:
+                raise MissingRequiredAttribute("Missing at_hash property",
+                                               idt)
+            if idt["at_hash"] != left_hash(msg["access_token"],
+                                           hfunc):
+                raise AtHashError(
+                    "Failed to verify access_token hash", idt)
+
+        if "code" in msg:
+            if "c_hash" not in idt:
+                raise MissingRequiredAttribute("Missing c_hash property",
+                                               idt)
+            if idt["c_hash"] != left_hash(msg["code"], hfunc):
+                raise CHashError("Failed to verify code hash", idt)
+
+    msg[verified_claim_name(claim)] = idt
+    logger.info('Verified {}: {}'.format(claim, idt.to_dict()))
+
+    return True
+
+
 class AccessTokenResponse(oauth2.AccessTokenResponse):
     c_param = oauth2.AccessTokenResponse.c_param.copy()
     c_param.update({"id_token": SINGLE_OPTIONAL_STRING})
 
     def verify(self, **kwargs):
         super(AccessTokenResponse, self).verify(**kwargs)
-        if "id_token" in self:
-            # Try to decode the JWT, checks the signature
-            args = {}
-            for arg in ID_TOKEN_VERIFY_ARGS:
-                try:
-                    args[arg] = kwargs[arg]
-                except KeyError:
-                    pass
-            idt = IdToken().from_jwt(str(self["id_token"]), **args)
-            if not idt.verify(**kwargs):
-                return False
 
-            self["__verified_id_token"] = idt
-            logger.info('Verified ID Token: {}'.format(idt.to_dict()))
+        clear_verified_claims(self)
+
+        if "id_token" in self:
+            if not verify_id_token(self, **kwargs):
+                return False
 
         return True
 
@@ -268,7 +353,7 @@ class AccessTokenResponse(oauth2.AccessTokenResponse):
 class UserInfoRequest(Message):
     c_param = {
         "access_token": SINGLE_OPTIONAL_STRING,
-    }
+        }
 
 
 class AuthorizationResponse(oauth2.AuthorizationResponse,
@@ -280,11 +365,14 @@ class AuthorizationResponse(oauth2.AuthorizationResponse,
         # "nonce": SINGLE_OPTIONAL_STRING,
         "access_token": SINGLE_OPTIONAL_STRING,
         "token_type": SINGLE_OPTIONAL_STRING,
-        "id_token": SINGLE_OPTIONAL_IDTOKEN
-    })
+        "id_token": SINGLE_OPTIONAL_IDTOKEN,
+        # Below is REQUIRED if doing session management
+        "session_state": SINGLE_OPTIONAL_STRING
+        })
 
     def verify(self, **kwargs):
         super(AuthorizationResponse, self).verify(**kwargs)
+        clear_verified_claims(self)
 
         if "aud" in self:
             if "client_id" in kwargs:
@@ -293,39 +381,9 @@ class AuthorizationResponse(oauth2.AuthorizationResponse,
                     return False
 
         if "id_token" in self:
-            # Try to decode the JWT, checks the signature
-            args = {}
-            for arg in ID_TOKEN_VERIFY_ARGS:
-                try:
-                    args[arg] = kwargs[arg]
-                except KeyError:
-                    pass
-            idt = IdToken().from_jwt(str(self["id_token"]), **args)
-            if not idt.verify(**kwargs):
-                raise VerificationError("Could not verify id_token", idt)
+            if not verify_id_token(self, check_hash=True, **kwargs):
+                return False
 
-            _alg = idt.jws_header["alg"]
-            # What if _alg == 'none'
-
-            hfunc = "HS" + _alg[-3:]
-
-            if "access_token" in self:
-                if "at_hash" not in idt:
-                    raise MissingRequiredAttribute("Missing at_hash property",
-                                                   idt)
-                if idt["at_hash"] != jws.left_hash(self["access_token"],
-                                                   hfunc):
-                    raise AtHashError(
-                        "Failed to verify access_token hash", idt)
-
-            if "code" in self:
-                if "c_hash" not in idt:
-                    raise MissingRequiredAttribute("Missing c_hash property",
-                                                   idt)
-                if idt["c_hash"] != jws.left_hash(self["code"], hfunc):
-                    raise CHashError("Failed to verify code hash", idt)
-
-            self["__verified_id_token"] = idt
         return True
 
 
@@ -351,7 +409,7 @@ class AuthorizationRequest(oauth2.AuthorizationRequest):
             "redirect_uri": SINGLE_REQUIRED_STRING,
             "nonce": SINGLE_OPTIONAL_STRING,
             "display": SINGLE_OPTIONAL_STRING,
-            "prompt": OPTIONAL_LIST_OF_STRINGS,
+            "prompt": OPTIONAL_LIST_OF_SP_SEP_STRINGS,
             "max_age": SINGLE_OPTIONAL_INT,
             "ui_locales": OPTIONAL_LIST_OF_SP_SEP_STRINGS,
             "claims_locales": OPTIONAL_LIST_OF_SP_SEP_STRINGS,
@@ -364,13 +422,13 @@ class AuthorizationRequest(oauth2.AuthorizationRequest):
             "request_uri": SINGLE_OPTIONAL_STRING,
             # "session_state": SINGLE_OPTIONAL_STRING,
             "response_mode": SINGLE_OPTIONAL_STRING,
-        }
-    )
+            }
+        )
     c_allowed_values = oauth2.AuthorizationRequest.c_allowed_values.copy()
     c_allowed_values.update({
         "display": ["page", "popup", "touch", "wap"],
         "prompt": ["none", "login", "consent", "select_account"]
-    })
+        })
 
     def verify(self, **kwargs):
         """Authorization Request parameters that are OPTIONAL in the OAuth 2.0
@@ -383,8 +441,11 @@ class AuthorizationRequest(oauth2.AuthorizationRequest):
         match."""
         super(AuthorizationRequest, self).verify(**kwargs)
 
+        clear_verified_claims(self)
+
         args = {}
-        for arg in ["keyjar", "opponent_id", "sender"]:
+        for arg in ["keyjar", "opponent_id", "sender", "alg", "encalg",
+                    "encenc"]:
             try:
                 args[arg] = kwargs[arg]
             except KeyError:
@@ -394,23 +455,32 @@ class AuthorizationRequest(oauth2.AuthorizationRequest):
             args["opponent_id"] = self["client_id"]
 
         if "request" in self:
-            if isinstance(self["request"], six.string_types):
+            if isinstance(self["request"], str):
                 # Try to decode the JWT, checks the signature
                 oidr = OpenIDRequest().from_jwt(str(self["request"]), **args)
 
-                # verify that nothing is change in the original message
+                # check if something is change in the original message
                 for key, val in oidr.items():
                     if key in self:
                         if self[key] != val:
-                            raise ValueError('{} != {}'.format(self[key], val))
+                            # log but otherwise ignore
+                            logger.warning('{} != {}'.format(self[key], val))
+
+                # remove all claims
+                _keys = list(self.keys())
+                for key in _keys:
+                    if key not in oidr:
+                        del self[key]
+
+                self.update(oidr)
 
                 # replace the JWT with the parsed and verified instance
-                self["request"] = oidr
+                self[verified_claim_name("request")] = oidr
 
         if "id_token_hint" in self:
-            if isinstance(self["id_token_hint"], six.string_types):
+            if isinstance(self["id_token_hint"], str):
                 idt = IdToken().from_jwt(str(self["id_token_hint"]), **args)
-                self["id_token_hint"] = idt
+                self["verified_id_token_hint"] = idt
 
         if "response_type" not in self:
             raise MissingRequiredAttribute("response_type missing", self)
@@ -445,22 +515,26 @@ class AuthorizationRequest(oauth2.AuthorizationRequest):
 
 class AccessTokenRequest(oauth2.AccessTokenRequest):
     c_param = oauth2.AccessTokenRequest.c_param.copy()
-    c_param.update({"client_assertion_type": SINGLE_OPTIONAL_STRING,
-                    "client_assertion": SINGLE_OPTIONAL_STRING})
+    c_param.update({
+        "client_assertion_type": SINGLE_OPTIONAL_STRING,
+        "client_assertion": SINGLE_OPTIONAL_STRING
+        })
     c_default = {"grant_type": "authorization_code"}
     c_allowed_values = {
         "client_assertion_type": [
             "urn:ietf:params:oauth:client-assertion-type:jwt-bearer"],
-    }
+        }
 
 
 class AddressClaim(Message):
-    c_param = {"formatted": SINGLE_OPTIONAL_STRING,
-               "street_address": SINGLE_OPTIONAL_STRING,
-               "locality": SINGLE_OPTIONAL_STRING,
-               "region": SINGLE_OPTIONAL_STRING,
-               "postal_code": SINGLE_OPTIONAL_STRING,
-               "country": SINGLE_OPTIONAL_STRING}
+    c_param = {
+        "formatted": SINGLE_OPTIONAL_STRING,
+        "street_address": SINGLE_OPTIONAL_STRING,
+        "locality": SINGLE_OPTIONAL_STRING,
+        "region": SINGLE_OPTIONAL_STRING,
+        "postal_code": SINGLE_OPTIONAL_STRING,
+        "country": SINGLE_OPTIONAL_STRING
+        }
 
 
 class OpenIDSchema(ResponseMessage):
@@ -489,7 +563,7 @@ class OpenIDSchema(ResponseMessage):
             "updated_at": SINGLE_OPTIONAL_INT,
             "_claim_names": OPTIONAL_MESSAGE,
             "_claim_sources": OPTIONAL_MESSAGE
-        })
+            })
 
     def verify(self, **kwargs):
         super(OpenIDSchema, self).verify(**kwargs)
@@ -549,10 +623,16 @@ class RegistrationRequest(Message):
         # "client_secret": SINGLE_OPTIONAL_STRING,
         # "access_token": SINGLE_OPTIONAL_STRING,
         "post_logout_redirect_uris": OPTIONAL_LIST_OF_STRINGS,
-    }
+        "frontchannel_logout_uri": SINGLE_OPTIONAL_STRING,
+        "frontchannel_logout_session_required": SINGLE_OPTIONAL_BOOLEAN,
+        "backchannel_logout_uri": SINGLE_OPTIONAL_STRING,
+        "backchannel_logout_session_supported": SINGLE_OPTIONAL_BOOLEAN
+        }
     c_default = {"application_type": "web", "response_types": ["code"]}
-    c_allowed_values = {"application_type": ["native", "web"],
-                        "subject_type": ["public", "pairwise"]}
+    c_allowed_values = {
+        "application_type": ["native", "web"],
+        "subject_type": ["public", "pairwise"]
+        }
 
     def verify(self, **kwargs):
         super(RegistrationRequest, self).verify(**kwargs)
@@ -595,7 +675,7 @@ class RegistrationResponse(ResponseMessage):
             "registration_client_uri": SINGLE_OPTIONAL_STRING,
             "client_id_issued_at": SINGLE_OPTIONAL_INT,
             "client_secret_expires_at": SINGLE_OPTIONAL_INT
-        })
+            })
     c_param.update(RegistrationRequest.c_param)
 
     def verify(self, **kwargs):
@@ -618,9 +698,11 @@ class RegistrationResponse(ResponseMessage):
 
 
 class ClientRegistrationErrorResponse(oauth2.ResponseMessage):
-    c_allowed_values = {"error": ["invalid_redirect_uri",
-                                  "invalid_client_metadata",
-                                  "invalid_configuration_parameter"]}
+    c_allowed_values = {
+        "error": ["invalid_redirect_uri",
+                  "invalid_client_metadata",
+                  "invalid_configuration_parameter"]
+        }
 
 
 class IdToken(OpenIDSchema):
@@ -638,8 +720,9 @@ class IdToken(OpenIDSchema):
         "acr": SINGLE_OPTIONAL_STRING,
         "amr": OPTIONAL_LIST_OF_STRINGS,
         "azp": SINGLE_OPTIONAL_STRING,
-        "sub_jwk": SINGLE_OPTIONAL_STRING
-    })
+        "sub_jwk": SINGLE_OPTIONAL_STRING,
+        "sid": SINGLE_OPTIONAL_STRING
+        })
     hashable = {'access_token': 'at_hash', 'code': 'c_hash'}
 
     def val_hash(self, alg):
@@ -647,7 +730,7 @@ class IdToken(OpenIDSchema):
 
         for attr, hash_attr in self.hashable.items():
             try:
-                self[hash_attr] = jws.left_hash(as_unicode(self[attr]), halg)
+                self[hash_attr] = left_hash(as_unicode(self[attr]), halg)
             except KeyError:
                 pass
             else:
@@ -666,13 +749,13 @@ class IdToken(OpenIDSchema):
         else:
             self.pack_init()
 
-        if 'jti' in self.c_param:
-            try:
-                _jti = kwargs['jti']
-            except KeyError:
-                _jti = uuid.uuid4().hex
-
-            self['jti'] = _jti
+        # if 'jti' in self.c_param:
+        #     try:
+        #         _jti = kwargs['jti']
+        #     except KeyError:
+        #         _jti = uuid.uuid4().hex
+        #
+        #     self['jti'] = _jti
 
     def to_jwt(self, key=None, algorithm="", lev=0, lifetime=0):
         self.pack(alg=algorithm, lifetime=lifetime)
@@ -753,87 +836,12 @@ class MessageWithIdToken(Message):
     def verify(self, **kwargs):
         super(MessageWithIdToken, self).verify(**kwargs)
         if "id_token" in self:
-            # Try to decode the JWT, checks the signature
-            args = {}
-            for arg in ID_TOKEN_VERIFY_ARGS:
-                try:
-                    args[arg] = kwargs[arg]
-                except KeyError:
-                    pass
-            idt = IdToken().from_jwt(str(self["id_token"]), **args)
-            if not idt.verify(**kwargs):
+            if not verify_id_token(self, **kwargs):
                 return False
 
-            # replace the JWT with the IdToken instance
-            self["id_token"] = idt
-
         return True
 
 
-class RefreshSessionRequest(MessageWithIdToken):
-    c_param = MessageWithIdToken.c_param.copy()
-    c_param.update({"redirect_url": SINGLE_REQUIRED_STRING,
-                    "state": SINGLE_REQUIRED_STRING})
-
-
-class RefreshSessionResponse(MessageWithIdToken, ResponseMessage):
-    c_param = MessageWithIdToken.c_param.copy()
-    c_param.update(ResponseMessage.c_param.copy())
-    c_param.update({"state": SINGLE_REQUIRED_STRING})
-
-
-class CheckSessionRequest(MessageWithIdToken):
-    pass
-
-
-class CheckIDRequest(Message):
-    c_param = {"access_token": SINGLE_REQUIRED_STRING}
-
-
-class EndSessionRequest(Message):
-    c_param = {
-        "id_token_hint": SINGLE_OPTIONAL_IDTOKEN,
-        "post_logout_redirect_uri": SINGLE_OPTIONAL_STRING,
-        "state": SINGLE_OPTIONAL_STRING
-    }
-
-    def verify(self, **kwargs):
-        super(EndSessionRequest, self).verify(**kwargs)
-        if "id_token_hint" in self:
-            # Try to decode the JWT, checks the signature
-            args = {}
-            for arg in ID_TOKEN_VERIFY_ARGS:
-                try:
-                    args[arg] = kwargs[arg]
-                except KeyError:
-                    pass
-            idt = IdToken().from_jwt(str(self["id_token_hint"]), **args)
-            if not idt.verify(**kwargs):
-                return False
-
-            # replace the JWT with the IdToken instance
-            self["id_token_hint"] = idt
-
-        return True
-
-
-class EndSessionResponse(ResponseMessage):
-    c_param = ResponseMessage.c_param.copy()
-    c_param.update({"state": SINGLE_OPTIONAL_STRING})
-
-
-class Claims(Message):
-    # c_param = {"*": SINGLE_OPTIONAL_JSON_CONV}
-    pass
-
-
-class ClaimsRequest(Message):
-    c_param = {
-        "userinfo": OPTIONAL_MULTIPLE_Claims,
-        "id_token": OPTIONAL_MULTIPLE_Claims
-    }
-
-
 class OpenIDRequest(AuthorizationRequest):
     pass
 
@@ -881,18 +889,24 @@ class ProviderConfigurationResponse(ResponseMessage):
         "op_tos_uri": SINGLE_OPTIONAL_STRING,
         "check_session_iframe": SINGLE_OPTIONAL_STRING,
         "end_session_endpoint": SINGLE_OPTIONAL_STRING,
+        "frontchannel_logout_supported": SINGLE_OPTIONAL_BOOLEAN,
+        "frontchannel_logout_session_supported": SINGLE_OPTIONAL_BOOLEAN,
+        "backchannel_logout_supported": SINGLE_OPTIONAL_BOOLEAN,
+        "backchannel_logout_session_supported": SINGLE_OPTIONAL_BOOLEAN
         # "jwk_encryption_url": SINGLE_OPTIONAL_STRING,
         # "x509_url": SINGLE_REQUIRED_STRING,
         # "x509_encryption_url": SINGLE_OPTIONAL_STRING,
-    })
-    c_default = {"version": "3.0",
-                 "token_endpoint_auth_methods_supported": [
-                     "client_secret_basic"],
-                 "claims_parameter_supported": False,
-                 "request_parameter_supported": False,
-                 "request_uri_parameter_supported": True,
-                 "require_request_uri_registration": True,
-                 "grant_types_supported": ["authorization_code", "implicit"]}
+        })
+    c_default = {
+        "version": "3.0",
+        "token_endpoint_auth_methods_supported": [
+            "client_secret_basic"],
+        "claims_parameter_supported": False,
+        "request_parameter_supported": False,
+        "request_uri_parameter_supported": True,
+        "require_request_uri_registration": True,
+        "grant_types_supported": ["authorization_code", "implicit"]
+        }
 
     def verify(self, **kwargs):
         super(ProviderConfigurationResponse, self).verify(**kwargs)
@@ -904,16 +918,21 @@ class ProviderConfigurationResponse(ResponseMessage):
                 check_char_set(scope, SCOPE_CHARSET)
 
         parts = urlparse(self["issuer"])
-        if parts.scheme != "https":
+        if 'allow_http' in kwargs:
+            pass
+        elif parts.scheme != "https":
             raise SchemeError("Not HTTPS")
 
+        if "RS256" not in self["id_token_signing_alg_values_supported"]:
+            raise ValueError('RS256 missing from id_token_signing_alg_values_supported')
+
         if not parts.query and not parts.fragment:
             pass
         else:
             raise ValueError('Issuer ID invalid')
 
         if any("code" in rt for rt in self[
-            "response_types_supported"]) and "token_endpoint" not in self:
+                "response_types_supported"]) and "token_endpoint" not in self:
             raise MissingRequiredAttribute("token_endpoint")
 
         return True
@@ -929,7 +948,7 @@ class JsonWebToken(Message):
         "nbf": SINGLE_OPTIONAL_INT,
         "iat": SINGLE_OPTIONAL_INT,
         "jti": SINGLE_OPTIONAL_STRING,
-    }
+        }
 
     def verify(self, **kwargs):
         super(JsonWebToken, self).verify(**kwargs)
@@ -949,6 +968,14 @@ class JsonWebToken(Message):
             if (_now - _skew) > _exp:
                 raise EXPError('Invalid expiration time')
 
+        try:
+            _iat = self['iat']
+        except KeyError:
+            pass
+        else:
+            if _iat > (_now + _skew):
+                raise EXPError('Invalid issued-at time')
+
         try:
             _nbf = self['nbf']
         except KeyError:
@@ -968,6 +995,10 @@ class JsonWebToken(Message):
             except KeyError:
                 pass
 
+        if 'iss' in kwargs and 'iss' in self:
+            if kwargs['iss'] != self['iss']:
+                raise ValueError('Wrong issuer')
+
         return True
 
 
@@ -979,14 +1010,14 @@ class AuthnToken(JsonWebToken):
         "jti": SINGLE_REQUIRED_STRING,
         "exp": SINGLE_REQUIRED_INT,
         "iat": SINGLE_OPTIONAL_INT,
-    }
+        }
 
 
 def jwt_deser(val, sformat="json"):
     if sformat == "urlencoded":
         sformat = "json"
     if sformat in ["dict", "json"]:
-        if not isinstance(val, six.string_types):
+        if not isinstance(val, str):
             val = json.dumps(val)
             sformat = "json"
     return JsonWebToken().deserialize(val, sformat)
@@ -996,13 +1027,17 @@ SINGLE_OPTIONAL_JWT = (Message, False, msg_ser, jwt_deser, False)
 
 
 class UserInfoErrorResponse(oauth2.ResponseMessage):
-    c_allowed_values = {"error": ["invalid_schema", "invalid_request",
-                                  "invalid_token", "insufficient_scope"]}
+    c_allowed_values = {
+        "error": ["invalid_schema", "invalid_request",
+                  "invalid_token", "insufficient_scope"]
+        }
 
 
 class DiscoveryRequest(Message):
-    c_param = {"resource": SINGLE_REQUIRED_STRING,
-               "rel": SINGLE_REQUIRED_STRING}
+    c_param = {
+        "resource": SINGLE_REQUIRED_STRING,
+        "rel": SINGLE_REQUIRED_STRING
+        }
 
 
 class Link(Message):
@@ -1015,7 +1050,7 @@ class Link(Message):
         "href": SINGLE_OPTIONAL_STRING,
         "titles": SINGLE_OPTIONAL_DICT,
         "properties": SINGLE_OPTIONAL_DICT
-    }
+        }
 
 
 def _l_deser(val, sformat):
@@ -1080,14 +1115,14 @@ class JRD(ResponseMessage):
         "aliases": OPTIONAL_LIST_OF_STRINGS,
         "properties": SINGLE_OPTIONAL_DICT,
         "links": REQUIRED_LINKS
-    }
+        }
 
 
 class WebFingerRequest(Message):
     c_param = {
         "resource": SINGLE_REQUIRED_STRING,
         "rel": SINGLE_REQUIRED_STRING
-    }
+        }
     c_default = {"rel": "http://openid.net/specs/connect/1.0/issuer"}
 
 
@@ -1095,28 +1130,16 @@ class ResourceRequest(Message):
     c_param = {"access_token": SINGLE_OPTIONAL_STRING}
 
 
-SCOPE2CLAIMS = {
-    "openid": ["sub"],
-    "profile": ["name", "given_name", "family_name", "middle_name",
-                "nickname", "profile", "picture", "website", "gender",
-                "birthdate", "zoneinfo", "locale", "updated_at",
-                "preferred_username"],
-    "email": ["email", "email_verified"],
-    "address": ["address"],
-    "phone": ["phone_number", "phone_number_verified"],
-    "offline_access": []
-}
+class Claims(Message):
+    # c_param = {"*": SINGLE_OPTIONAL_JSON_CONV}
+    pass
 
 
-def scope2claims(scopes):
-    res = {}
-    for scope in scopes:
-        try:
-            claims = dict([(name, None) for name in SCOPE2CLAIMS[scope]])
-            res.update(claims)
-        except KeyError:
-            continue
-    return res
+class ClaimsRequest(Message):
+    c_param = {
+        "userinfo": OPTIONAL_MULTIPLE_Claims,
+        "id_token": OPTIONAL_MULTIPLE_Claims
+        }
 
 
 def factory(msgtype, **kwargs):
@@ -1132,48 +1155,22 @@ def factory(msgtype, **kwargs):
     return oauth2.factory(msgtype, **kwargs)
 
 
-def make_openid_request(arq, keys=None, userinfo_claims=None,
-                        idtoken_claims=None, request_object_signing_alg=None,
-                        **kwargs):
+def make_openid_request(arq, keys, issuer, request_object_signing_alg, recv):
     """
     Construct the JWT to be passed by value (the request parameter) or by
     reference (request_uri).
     The request will be signed
 
     :param arq: The Authorization request
-    :param keys: Keys to use for signing/encrypting
-    :param userinfo_claims: UserInfo claims
-    :param idtoken_claims: IdToken claims
+    :param keys: Keys to use for signing/encrypting. A KeyJar instance
+    :param issuer: Who is signing this JSON Web Token
     :param request_object_signing_alg: Which signing algorithm to use
+    :param recv: The intended receiver of the request
     :return: JWT encoded OpenID request
     """
 
-    oir_args = {}
-    for prop in OpenIDRequest.c_param.keys():
-        try:
-            oir_args[prop] = arq[prop]
-        except KeyError:
-            pass
-
-    for attr in ["scope", "response_type"]:
-        if attr in oir_args:
-            oir_args[attr] = " ".join(oir_args[attr])
-
-    c_args = {}
-    if userinfo_claims is not None:
-        # UserInfoClaims
-        c_args["userinfo"] = Claims(**userinfo_claims)
-
-    if idtoken_claims is not None:
-        # IdTokenClaims
-        c_args["id_token"] = Claims(**idtoken_claims)
-
-    if c_args:
-        oir_args["claims"] = ClaimsRequest(**c_args)
-
-    oir = OpenIDRequest(**oir_args)
-
-    return oir.to_jwt(key=keys, algorithm=request_object_signing_alg)
+    _jwt = JWT(key_jar=keys, iss=issuer, sign_alg=request_object_signing_alg)
+    return _jwt.pack(arq.to_dict(), owner=issuer, recv=recv)
 
 
 def claims_match(value, claimspec):
@@ -1181,10 +1178,10 @@ def claims_match(value, claimspec):
     Implements matching according to section 5.5.1 of
     http://openid.net/specs/openid-connect-core-1_0.html
     The lack of value is not checked here.
-    Also the text doesn't prohibit claims specification having both 'value' 
+    Also the text doesn't prohibit claims specification having both 'value'
     and 'values'.
 
-    :param value: single value or list of values
+    :param value: single value
     :param claimspec: None or a dictionary with 'essential', 'value' or 'values'
         as keys
     :return: Boolean
@@ -1207,8 +1204,8 @@ def claims_match(value, claimspec):
         if matched:
             break
 
-    if matched is False:
-        if list(claimspec.keys()) == ['essential']:
-            return True
+    # No values to test against so it's just about being there or not
+    if list(claimspec.keys()) == ['essential']:
+        return True
 
     return matched
